# -- Imports ------------------------------------------------------------------

from collections.abc import Iterator, Mapping
from math import log2
from typing import List, Optional, Union

# -- Class --------------------------------------------------------------------


class ADCConfiguration(Mapping):
    """Support for reading and writing the ADC configuration"""

    def __init__(
        self,
        *data: Union[bytearray, List[int]],
        set: Optional[bool] = None,
        prescaler: Optional[int] = None,
        acquisition_time: Optional[int] = None,
        oversampling_rate: Optional[int] = None,
        reference_voltage: Optional[float] = None,
    ):
        """Initialize the ADC configuration using the given arguments

        Positional Parameters
        ---------------------

        data:
            A list containing the (first five) bytes of the ADC configuration

        Keyword Parameters
        ------------------

        set:
            Specifies if we want to set or retrieve (get) the ADC configuration

        prescaler:
            The ADC prescaler value (1 – 127)

        acquisition_time:
            The acquisition time in number of cycles
            (1, 2, 3, 4, 8, 16, 32, … , 256)

        oversampling_rate:
            The ADC oversampling rate (1, 2, 4, 8, … , 4096)

        reference_voltage:
            The ADC reference voltage in Volt
            (1.25, 1.65, 1.8, 2.1, 2.2, 2.5, 2.7, 3.3, 5, 6.6)

        """

        if data:
            data_bytes = list(data[0])
            if not isinstance(data_bytes, list):
                raise ValueError(
                    "Unsupported object type for argument data: "
                    f"“{type(data_bytes)}”"
                )
            required_length = 5
            if len(data_bytes) < required_length:
                raise ValueError(
                    f"Data length of {len(data_bytes)} is too "
                    "small, at least length of "
                    f"“{required_length}” required"
                )
            self.data = data_bytes[0:5] + [0] * 3
        else:
            self.data = [0] * 8

        # ==================
        # = Get/Set Config =
        # ==================

        if set is not None:
            get_set_byte = self.data[0]
            # Set get/set to 0
            get_set_byte &= 0b01111111
            # Set value
            get_set_byte |= int(set) << 7
            self.data[0] = get_set_byte

        # =============
        # = Prescaler =
        # =============

        if prescaler is not None:
            if not (1 <= prescaler <= 127):
                raise ValueError(
                    f"Prescaler value of “{prescaler}” out of range"
                    ", please use a value between 1 and 127"
                )
            self.data[1] = prescaler
        elif self.data[1] == 0:
            # Make sure default prescaler value makes sense
            self.data[1] = 8

        # ====================
        # = Acquisition Time =
        # ====================

        if acquisition_time is not None:
            possible_acquisition_times = list(range(1, 4)) + [
                2**value for value in range(3, 9)
            ]
            if acquisition_time not in possible_acquisition_times:
                raise ValueError(
                    f"Acquisition time of “{acquisition_time}” out of range"
                    ", please use one of the following values: "
                    + ", ".join(map(str, possible_acquisition_times))
                )

            acquisition_time_byte = (
                acquisition_time - 1
                if acquisition_time <= 3
                else int(log2(acquisition_time)) + 1
            )

            self.data[2] = acquisition_time_byte

        # =====================
        # = Oversampling Rate =
        # =====================

        if oversampling_rate is not None:
            possible_oversampling_rates = [2**value for value in range(13)]
            if oversampling_rate not in possible_oversampling_rates:
                raise ValueError(
                    f"Oversampling rate of “{oversampling_rate}” out of"
                    "range, please use one of the following values: "
                    + ", ".join(map(str, possible_oversampling_rates))
                )

            self.data[3] = int(log2(oversampling_rate))

        # =====================
        # = Reference Voltage =
        # =====================

        if reference_voltage is not None:
            possible_reference_voltages = (
                1.25,
                1.65,
                1.8,
                2.1,
                2.2,
                2.5,
                2.7,
                3.3,
                5,
                6.6,
            )
            if reference_voltage not in possible_reference_voltages:
                raise ValueError(
                    f"Reference voltage of “{oversampling_rate}” out of range"
                    ", please use one of the following values: "
                    + ", ".join(map(str, possible_reference_voltages))
                )

            self.data[4] = int(reference_voltage * 20)
        elif self.data[4] == 0:
            # Make sure default reference voltage value makes sense
            supply_voltage = 3.3
            self.data[4] = int(supply_voltage * 20)

        self.attributes = {
            "reference_voltage": self.reference_voltage,
            "prescaler": self.prescaler,
            "acquisition_time": self.acquisition_time,
            "oversampling_rate": self.oversampling_rate,
        }

    def __getitem__(self, item: str) -> float:
        """Return values of the mapping provided by this class

        Note: This method allow access to the object via the splat
              operators (*, **)

        Parameters
        ----------

        item:
            The attribute for which we want to retrieve the value

        Returns
        -------

        The value of the attribute

        Examples
        --------

        >>> dict(**ADCConfiguration()) # doctest:+NORMALIZE_WHITESPACE
        {'reference_voltage': 3.3,
         'prescaler': 8,
         'acquisition_time': 1,
         'oversampling_rate': 1}

        >>> dict(**ADCConfiguration(oversampling_rate=64)
        ...     ) # doctest:+NORMALIZE_WHITESPACE
        {'reference_voltage': 3.3,
         'prescaler': 8,
         'acquisition_time': 1,
         'oversampling_rate': 64}

        """

        return self.attributes[item]()

    def __iter__(self) -> Iterator:
        """Return an iterator over the mapping provided by this class

        Note: This method allow access to the object via the splat
              operators (*, **)

        Returns
        -------

        The names of the “important” properties of the ADC configuration:

        - reference voltage
        - prescaler
        - acquisition time
        - oversampling rate

        Examples
        --------

        >>> for attribute in ADCConfiguration():
        ...     print(attribute)
        reference_voltage
        prescaler
        acquisition_time
        oversampling_rate

        """

        return iter(self.attributes)

    def __len__(self) -> int:
        """Return the length of the mapping provided by this class

        Note: This method allow access to the object via the splat
              operators (*, **)

        Returns
        -------

        The amount of the “important” properties of the ADC configuration:

        - reference voltage
        - prescaler
        - acquisition time
        - oversampling rate

        Examples
        --------

        >>> len(ADCConfiguration())
        4

        >>> len(ADCConfiguration(reference_voltage=3.3))
        4

        """

        return len(self.attributes)

    def __repr__(self) -> str:
        """Retrieve the textual representation of the ADC configuration

        Returns
        -------

        A string that describes the ADC configuration

        Examples
        --------

        >>> ADCConfiguration(prescaler=1, reference_voltage=3.3
        ... ) # doctest:+NORMALIZE_WHITESPACE
        Get, Prescaler: 1, Acquisition Time: 1, Oversampling Rate: 1,
        Reference Voltage: 3.3 V

        >>> ADCConfiguration(
        ...     set=True,
        ...     prescaler=64,
        ...     acquisition_time=128,
        ...     oversampling_rate=1024,
        ...     reference_voltage=1.8) # doctest:+NORMALIZE_WHITESPACE
        Set, Prescaler: 64, Acquisition Time: 128, Oversampling Rate: 1024,
        Reference Voltage: 1.8 V

        >>> ADCConfiguration([0, 2, 4, 6, 25]) # doctest:+NORMALIZE_WHITESPACE
        Get, Prescaler: 2, Acquisition Time: 8, Oversampling Rate: 64,
        Reference Voltage: 1.25 V

        """

        set = bool(self.data[0] >> 7)
        prescaler = self.data[1]
        acquisition_time = (
            self.data[2] + 1 if self.data[2] <= 3 else 2 ** (self.data[2] - 1)
        )
        oversampling_rate = 2 ** self.data[3]
        reference_voltage = self.reference_voltage()

        parts = [
            "Set" if set else "Get",
            f"Prescaler: {prescaler}",
            f"Acquisition Time: {acquisition_time}",
            f"Oversampling Rate: {oversampling_rate}",
            f"Reference Voltage: {reference_voltage} V",
        ]

        return ", ".join(parts)

    def reference_voltage(self) -> float:
        """Get the reference voltage

        Returns
        -------

        The reference voltage in Volt

        Examples
        --------

        >>> ADCConfiguration(reference_voltage=3.3).reference_voltage()
        3.3

        >>> ADCConfiguration(reference_voltage=6.6).reference_voltage()
        6.6

        >>> ADCConfiguration(reference_voltage=1.8).reference_voltage()
        1.8

        """

        return self.data[4] / 20

    def prescaler(self) -> int:
        """Get the prescaler value

        Returns
        -------

        The prescaler value

        Examples
        --------

        >>> ADCConfiguration(prescaler=127).prescaler()
        127

        """

        return self.data[1]

    def acquisition_time(self) -> int:
        """Get the acquisition time

        Returns
        -------

        The acquisition time

        Examples
        --------

        >>> ADCConfiguration(acquisition_time=2).acquisition_time()
        2

        """

        acquisition_time_byte = self.data[2]

        return (
            acquisition_time_byte + 1
            if acquisition_time_byte <= 3
            else 2 ** (acquisition_time_byte - 1)
        )

    def oversampling_rate(self) -> int:
        """Get the oversampling rate

        Returns
        -------

        The oversampling rate

        Examples
        --------

        >>> ADCConfiguration(oversampling_rate=128).oversampling_rate()
        128

        """

        oversampling_rate_byte = self.data[3]

        return 2**oversampling_rate_byte

    def sample_rate(self) -> int:
        """Calculate the sampling rate for the current ADC configuration

        Returns
        -------

        The calculated sample rate

        Examples
        --------

        >>> ADCConfiguration(prescaler=2, acquisition_time=8,
        ...                  oversampling_rate=64).sample_rate()
        9524

        >>> ADCConfiguration(prescaler=8, acquisition_time=8,
        ...                  oversampling_rate=64).sample_rate()
        3175

        >>> ADCConfiguration(reference_voltage=5.0,
        ...                  prescaler=16,
        ...                  acquisition_time=8,
        ...                  oversampling_rate=128).sample_rate()
        840

        """

        clock_frequency = 38_400_000

        return round(
            clock_frequency
            / (
                (self.prescaler() + 1)
                * (self.acquisition_time() + 13)
                * self.oversampling_rate()
            )
        )


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from pint import Quantity

from mytoolit.measurement.constants import ADC_MAX_VALUE
from mytoolit.measurement.units import volt

# -- Functions ----------------------------------------------------------------


def convert_raw_to_supply_voltage(
    voltage_raw: int, reference_voltage: float = 3.3
) -> Quantity:
    """Convert a raw 2 byte supply voltage value to a voltage value

    Parameters
    ----------

    voltage_raw:
        A 2 byte supply voltage value (returned by the `Streaming` command)

    reference_voltage:
        The ADC reference voltage

    Returns
    -------

    The supply voltage in volts

    Example:

    >>> voltage = convert_raw_to_supply_voltage(11000)
    >>> volt(3.15) < voltage < volt(3.16)
    True

    >>> voltage = convert_raw_to_supply_voltage(2**15, reference_voltage=1.8)
    >>> volt(5.12) < voltage < volt(5.14)
    True

    """

    if voltage_raw <= 0:
        return volt(0)

    # The value below is the result of the voltage divider circuit, which
    # uses a 4.7 kΩ and 1 kΩ resistor: (470 kΩ + 100 kΩ) / 100 kΩ = 5.7
    # If you want to know why we used these resistor values, then please
    # take a look
    # [here](https://en.wikipedia.org/wiki/E_series_of_preferred_numbers).
    voltage_divider_factor = 5.7
    return volt(
        voltage_raw
        * voltage_divider_factor
        * reference_voltage
        / ADC_MAX_VALUE
    )


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from os import environ, pathsep
from pathlib import Path
from platform import system
from re import compile
from subprocess import run
from sys import byteorder
from typing import List, Optional, Union

from mytoolit.config import settings

# -- Classes ------------------------------------------------------------------


class CommanderException(Exception):
    """Describes problems regarding the execution of Simplicity Commander"""


class CommanderReturnCodeException(CommanderException):
    """A Simplicity Commander command did not return the success status code"""


class CommanderOutputMatchException(CommanderException):
    """The Simplicity Commander output did not contain an expected value"""


class Commander:
    """Wrapper for the Simplicity Commander commandline tool"""

    def __init__(self, serial_number: int, chip: str):
        """Initialize the Simplicity Commander wrapper

        serial_number:
            The serial number of the programming board that is connected to
            the hardware

        chip:
            The identifier of the chip on the PCB e.g. “BGM121A256V2”

        """

        self._add_path_to_environment()
        self.identification_arguments = [
            "--serialno",
            f"{serial_number}",
            "-d",
            chip,
        ]
        self.error_reasons = {
            "incorrect serial": (
                f"Serial number of programming board “{serial_number}”"
                " incorrect"
            ),
            "programmer not connected": (
                "Programming board is not connected to computer"
            ),
            "device not connected": (
                "Programming board is not connected to device"
            ),
        }

    def _add_path_to_environment(self) -> None:
        """Add path to Simplicity Commander (`commander`) to `PATH`

        After calling this method you should be able to call `commander`
        without its path prefix, if you installed it somewhere in the
        locations specified below `COMMANDS` → `PATH` in the configuration.

        Example
        -------

        >>> commander = Commander(
        ...     serial_number=settings.sth.programming_board.serial_number,
        ...     chip='BGM121A256V2')

        >>> from subprocess import run
        >>> result = run("commander --version".split(), capture_output=True)
        >>> result.returncode == 0
        True

        """

        path = settings.commands.path
        operating_system = system()
        paths = (
            path.linux
            if operating_system == "Linux"
            else path.mac if operating_system == "Darwin" else path.windows
        )

        environ["PATH"] += pathsep + pathsep.join(paths)

    def _run_command(
        self,
        command: List[str],
        description: str,
        possible_error_reasons: Optional[List[str]] = None,
        regex_output: Optional[str] = None,
    ) -> str:
        """Run a Simplicity Commander command

        Parameters
        ----------

        command:
            The Simplicity Commander subcommand including all necessary
            arguments

        description:
            A textual description of the purpose of the command
            e.g. “enable debug mode”

        possible_error_reasons:
            A list of dictionary keys that describe why the command might have
            failed

        regex_output:
            An optional regular expression that has to match part of the
            standard output of the command

        Raises
        ------

        A `CommanderException` if

        - the command returned unsuccessfully or
        - the standard output did not match the optional regular expression
          specified in `regex_output`

        Returns
        -------

        The standard output of the command

        """

        if possible_error_reasons:
            for reason in possible_error_reasons:
                if reason not in self.error_reasons:
                    raise ValueError(
                        f"“{reason}” is not a valid possible error reason"
                    )

        result = run(["commander"] + command, capture_output=True, text=True)

        if result.returncode != 0:
            # Since Windows seems to return the exit code as unsigned number we
            # need to convert it first to the “real” signed number.
            returncode = (
                int.from_bytes(
                    result.returncode.to_bytes(4, byteorder),
                    byteorder,
                    signed=True,
                )
                if system() == "Windows"
                else result.returncode
            )
            error_message = (
                "Execution of Simplicity Commander command to "
                f"{description} failed with return code "
                f"“{returncode}”"
            )
            combined_output = (
                "\n".join((result.stdout, result.stderr))
                if result.stdout or result.stderr
                else ""
            )
            if combined_output:
                error_message += (
                    "\n\nSimplicity Commander output:\n\n"
                    f"{combined_output.rstrip()}"
                )

            if possible_error_reasons:
                error_reasons = "\n".join(
                    [
                        f"• {self.error_reasons[reason]}"
                        for reason in possible_error_reasons
                    ]
                )
                error_message += (
                    f"\n\nPossible error reasons:\n\n{error_reasons}"
                )

            raise CommanderReturnCodeException(error_message)

        if (
            regex_output is not None
            and compile(regex_output).search(result.stdout) is None
        ):
            error_message = (
                "Output of Simplicity Commander command to "
                f"{description}:\n{result.stdout}\n"
                "did not match the expected regular expression "
                f"“{regex_output}”"
            )
            raise CommanderOutputMatchException(error_message)

        return result.stdout

    def enable_debug_mode(self) -> None:
        """Enable debug mode for external device

        Example
        -------

        Enable debug mode of STH programming board

        >>> from mytoolit.config import settings

        >>> commander = Commander(
        ...     serial_number=settings.sth.programming_board.serial_number,
        ...     chip='BGM121A256V2')
        >>> commander.enable_debug_mode()

        """

        error_reasons = ["programmer not connected", "incorrect serial"]
        self._run_command(
            command="adapter dbgmode OUT".split()
            + self.identification_arguments,
            description="enable debug mode",
            possible_error_reasons=error_reasons,
            regex_output="Setting debug mode to OUT",
        )

    def unlock_device(self) -> None:
        """Unlock device for debugging

        Calling this method will erase the flash of the device!

        """

        self._run_command(
            command="device unlock".split() + self.identification_arguments,
            description="unlock device",
            possible_error_reasons=[
                "device not connected",
                "programmer not connected",
                "incorrect serial",
            ],
            regex_output="Chip successfully unlocked",
        )

    def upload_flash(self, filepath: Union[str, Path]) -> None:
        """Upload code into the flash memory of the device

        Parameters
        ----------

        filepath:
            The filepath of the flash image

        """

        # Set debug mode to out, to make sure we flash the STH (connected via
        # debug cable) and not another microcontroller connected to the
        # programmer board.
        self.enable_debug_mode()

        # Unlock device (triggers flash erase)
        self.unlock_device()

        self._run_command(
            command=["flash", f"{filepath}", "--address", "0x0"]
            + self.identification_arguments,
            description="upload firmware",
        )

    def read_power_usage(self, milliseconds: float = 1000) -> float:
        """Read the power usage of the connected hardware

        Parameters
        ----------

        milliseconds:
            The amount of time the power usage should be measured for

        Returns
        -------

        The measured power usage in milliwatts

        Example
        -------

        Measure power usage of connected STH

        >>> from mytoolit.config import settings

        >>> commander = Commander(
        ...     serial_number=settings.sth.programming_board.serial_number,
        ...     chip='BGM121A256V2')
        >>> commander.read_power_usage() > 0
        True

        """

        command = [
            "aem",
            "measure",
            "--windowlength",
            str(milliseconds),
        ] + self.identification_arguments

        regex = r"Power\s*\[mW\]\s*:\s*(?P<milliwatts>\d+\.\d+)"
        try:
            output = self._run_command(
                command=command,
                description="read power usage",
                possible_error_reasons=[
                    "programmer not connected",
                    "incorrect serial",
                ],
                regex_output=regex,
            )
        except CommanderOutputMatchException:
            raise CommanderOutputMatchException(
                "Unable to extract power usage "
                "from Simplicity Commander output"
            )

        pattern_match = compile(regex).search(output)
        assert pattern_match is not None
        milliwatts = pattern_match["milliwatts"]

        return float(milliwatts)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from pint import Quantity, Unit, UnitRegistry

# -- Attributes ---------------------------------------------------------------

units = UnitRegistry()
del UnitRegistry

# -- Functions ----------------------------------------------------------------


def celsius(temperature: float) -> Quantity:
    """Return given value as temperature quantity

    Parameters
    ----------

    temperature:
        Magnitude of the temperature in degree Celsius

    Returns
    -------

    A temperature quantity with the given magnitude

    """

    # As workaround for the `UndefinedUnitError` the code below throws, if
    # we assert the type of `units.celsius` directly, we just use a
    # variable to access and check the type.
    _celsius = units.celsius
    assert isinstance(_celsius, Unit)

    return Quantity(temperature, _celsius)


def g0(acceleration: float) -> Quantity:
    """Return given value as acceleration quantity

    Parameters
    ----------

    acceleration:
        Magnitude of acceleration in multiples of g₀

    Returns
    -------

    An acceleration quantity with the given magnitude

    """

    _g0 = units.g0
    assert isinstance(_g0, Unit)

    return Quantity(acceleration, _g0)


def volt(voltage: float) -> Quantity:
    """Return given value as voltage quantity

    Parameters
    ----------

    voltage:
        Magnitude of voltage

    Returns
    -------

    A voltage quantity with the given magnitude

    """

    _volt = units.volt
    assert isinstance(_volt, Unit)

    return Quantity(voltage, _volt)



# -- Imports ------------------------------------------------------------------

from pathlib import Path
from platform import system
from subprocess import CalledProcessError, run
from typing import Union

# -- Classes ------------------------------------------------------------------


class UnableToOpenError(Exception):
    """Exception for errors that occur when opening a file in default app"""


# -- Functions ----------------------------------------------------------------


def open_file_windows(filepath: Union[Path, str]) -> None:
    """Open the given file in the default application on Windows

    Arguments
    ---------

    filepath:
        Location of the file that should be opened

    Raises
    ------

    An `UnableToOpenError` if opening the file failed

    """

    from os import startfile  # type: ignore[attr-defined]

    try:
        startfile(filepath)
    except FileNotFoundError as error:
        raise UnableToOpenError(f"File “{filepath}” does not exist") from error


def open_file_other(filepath: Union[Path, str]) -> None:
    """Open the given file in the default application on a non-Windows OS

    Arguments
    ---------

    filepath:
        Location of the file that should be opened

    Raises
    ------

    An `UnableToOpenError` if opening the file failed

    """

    try:
        command = "/usr/bin/open" if system() == "Darwin" else "xdg-open"
        run([command, filepath], capture_output=True, check=True, text=True)

    except CalledProcessError as error:
        raise UnableToOpenError(f"{error.stderr}") from error
    except FileNotFoundError as error:
        raise UnableToOpenError(
            f"Command “{command}” does not exist"
        ) from error


def open_file(filepath: Union[Path, str]) -> None:
    """Open the given file in the default application

    Arguments
    ---------

    filepath:
        Location of the file that should be opened

    Raises
    ------

    An `UnableToOpenError` if opening the file failed

    Examples
    --------

    Open (hopefully) non-existing file

    >>> try:
    ...     open_file("does-not-exist.txt")
    ... except UnableToOpenError:
    ...     print("UnableToOpenError")
    UnableToOpenError

    Open existing file

    >>> from importlib.resources import as_file, files
    >>> from platform import system
    >>> from os import environ
    >>> def test():
    ...     with as_file(
    ...         files("mytoolit.config").joinpath("config.yaml")
    ...     ) as repo_config_file:
    ...         open_file(repo_config_file)

    We only run the test in a graphical environments on Linux, since
    otherwise `xdg-open` will block while opening a CLI util to open the
    YAML config file.

    >>> if system() != "Linux" or environ.get("DISPLAY"):
    ...     test()

    """

    # Some tool (e.g. `xdg-open` in WSL) do not fail, even when we try to open
    # a non-existent file. We add this check to work around this behavior.
    if not Path(filepath).exists():
        raise UnableToOpenError(f"File “{filepath}” does not exist")

    if system() == "Windows":
        open_file_windows(filepath)
    else:
        open_file_other(filepath)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# Original Source: https://stackoverflow.com/questions/31712386

# -- Imports ------------------------------------------------------------------

from pdfrw import PdfReader, PdfDict
from pdfrw.buildxobj import pagexobj
from pdfrw.toreportlab import makerl
from reportlab.platypus import Flowable

# -- Class --------------------------------------------------------------------


class PDFImage(Flowable):
    """Create a flowable object from a PDF"""

    def __init__(self, filepath, width=200, height=200):
        """Initialize the flowable

        Parameters
        ----------

        filepath:
            A path to the PDF object this object represents
        width:
            The width of the PDF object
        height:
            The height of the PDF object
        """

        super().__init__()

        self.width = width
        self.height = height

        with open(filepath, "rb") as pdf_data:
            (page,) = PdfReader(pdf_data).pages
            self.data = pagexobj(page)

    def wrap(self, width, height):
        """Wrap the PDF object according to the given dimensions

        Parameters
        ----------

        width:
            The width of the wrapped PDF object
        height:
            The height of the wrapped PDF object

        Returns
        -------

        The dimensions of the wrapped object
        """
        return self.width, self.height

    def drawOn(self, canvas, x, y, _=0):
        """Draw the PDF object on the given canvas

        Parameters
        ----------

        canvas:
            The canvas where the flowable should be placed
        x:
            The x position of the PDF inside the canvas
        y:
            The y position of the PDF inside the canvas
        _:
            Ignored parameter that stores offset data for the alignment
        """

        canvas.saveState()

        data = self.data
        if isinstance(data, PdfDict):
            x_scale = self.width / data.BBox[2]
            y_scale = self.height / data.BBox[3]
            canvas.translate(x, y)
            canvas.scale(x_scale, y_scale)
            canvas.doForm(makerl(canvas, data))
        else:
            canvas.drawImage(self.img_data, x, y, self.width, self.height)

        canvas.restoreState()



# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from typing import Union

# -- Class --------------------------------------------------------------------


class EEPROMStatus:
    """This class represents an EEPROM status byte"""

    def __init__(self, status: Union[int, str, EEPROMStatus]) -> None:
        """Create a new status byte from the given argument

        Parameters
        ----------

        status:
            The value of the status byte

        Examples
        --------

        >>> EEPROMStatus(0xca)
        Locked (0xca)

        >>> EEPROMStatus(0)
        Uninitialized (0x00)

        >>> EEPROMStatus('Initialized')
        Initialized (0xac)

        >>> EEPROMStatus('Uninitialized')
        Uninitialized (0x00)

        >>> EEPROMStatus('Locked').value == 0xca
        True

        >>> EEPROMStatus(EEPROMStatus('Locked'))
        Locked (0xca)

        >>> EEPROMStatus('Something')
        Traceback (most recent call last):
           ...
        ValueError: Unknown EEPROM status “Something”

        """

        if isinstance(status, str):
            if status == "Initialized":
                self.value = 0xAC
                return

            if status == "Locked":
                self.value = 0xCA
                return

            if status == "Uninitialized":
                self.value = 0
                return

            raise ValueError(f"Unknown EEPROM status “{status}”")

        if isinstance(status, EEPROMStatus):
            self.value = status.value
            return

        self.value = status

    def __repr__(self) -> str:
        """Return the string representation of the status byte

        Returns
        -------

        A string that describes the current value of the status byte

        Examples
        --------

        >>> EEPROMStatus(0xac)
        Initialized (0xac)

        >>> EEPROMStatus(0x13)
        Uninitialized (0x13)

        >>> EEPROMStatus(0xCA)
        Locked (0xca)

        """

        value = self.value

        description = (
            "Initialized"
            if self.is_initialized()
            else ("Locked" if self.is_locked() else "Uninitialized")
        )

        return f"{description} (0x{value:02x})"

    def is_locked(self) -> bool:
        """Check if the EEPROM is locked

        Returns
        -------

        `True` if the status byte represents a locked EEPROM or `False`
        otherwise

        Examples
        --------

        >>> EEPROMStatus(0xca).is_locked()
        True

        >>> EEPROMStatus(0xac).is_locked()
        False

        >>> EEPROMStatus(0x2).is_locked()
        False

        """

        return self.value == 0xCA

    def is_initialized(self) -> bool:
        """Check if the EEPROM is initialized

        Returns
        -------

        `True` if the status byte represents an initialized EEPROM or `False`
        otherwise

        Examples
        --------

        >>> EEPROMStatus(0xca).is_initialized()
        False

        >>> EEPROMStatus(0xac).is_initialized()
        True

        >>> EEPROMStatus(0x2).is_initialized()
        False

        """

        return self.value == 0xAC


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Classes ------------------------------------------------------------------


class SensorConfig:
    """Used to store the configuration of the three sensor channels"""

    def __init__(self, first: int = 0, second: int = 0, third: int = 0):
        """Initialize the sensor configuration using the given arguments

        Parameters
        ----------

        first:
            The sensor number for the first measurement channel

        second:
            The sensor number for the second measurement channel

        third:
            The sensor number for the third measurement channel

        """

        self.first = first
        self.second = second
        self.third = third

    def __str__(self) -> str:
        """The string representation of the sensor configuration

        Returns
        -------

        A textual representation of the sensor configuration

        Examples
        --------

        >>> str(SensorConfig(first=1, second=3, third=2))
        'M1: S1, M2: S3, M3: S2'

        >>> str(SensorConfig())
        ''

        >>> str(SensorConfig(second=1))
        'M2: S1'

        """

        return ", ".join(
            (
                f"M{sensor}: S{value}"
                for sensor, value in enumerate(
                    (self.first, self.second, self.third), start=1
                )
                if value != 0
            )
        )

    def __repr__(self) -> str:
        """The textual representation of the sensor configuration

        Returns
        -------

        A textual representation of the sensor configuration

        Examples
        --------

        >>> repr(SensorConfig(first=1, second=3, third=2))
        'M1: S1, M2: S3, M3: S2'

        >>> repr(SensorConfig())
        'M1: None, M2: None, M3: None'

        """

        return ", ".join(
            (
                f"M{sensor}: {f'S{value}' if value != 0 else 'None'}"
                for sensor, value in enumerate(
                    (self.first, self.second, self.third), start=1
                )
            )
        )

    def disable_channel(
        self, first: bool = False, second: bool = False, third: bool = False
    ):
        """Disable certain (measurement) channels

        Parameters
        ----------

        first:
            Specifies if the first measurement channel should be disabled or
            not

        second:
            Specifies if the second measurement channel should be disabled or
            not

        third:
            Specifies if the third measurement channel should be disabled or
            not

        """

        if first:
            self.first = 0
        if second:
            self.second = 0
        if third:
            self.third = 0


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from asyncio import new_event_loop, set_event_loop, sleep as async_sleep
from datetime import date, datetime
from pathlib import Path
from types import SimpleNamespace
from typing import List, Union
from unittest import TestCase

from dynaconf.utils.boxing import DynaBox
from semantic_version import Version

from mytoolit import __version__
from mytoolit.can import Network, Node, State
from mytoolit.cmdline.commander import Commander
from mytoolit.config import settings
from mytoolit.eeprom import EEPROMStatus
from mytoolit.report import Report

# -- Class --------------------------------------------------------------------


class TestNode(TestCase):
    """This class contains shared test code for STH and STU

    You are not supposed to use this class directly. Instead use it as base
    class for your test class.

    Every subclass of this class has to implement the method `_read_data`,
    which sets the **class** attributes:

    - bluetooth_mac
    - bluetooth_rssi
    - firmware_version

    The method `_read_data` will be called after connection took place (the
    method `_connect` has been called). Please note, that this class only
    connects to the STU. If you also want to connect to a sensor node, please
    overwrite the method `_connect`.

    To add additional test attributes shown in the standard output and
    optionally the PDF, add them as **class** variables to the subclass. Then
    use the **class** method `add_attribute` in the method `setUpClass` and
    use a format string where you reference the class variable as value
    argument. Please do not forget to call `setUpClass` of the superclass
    before you do that.

    The various `_test` methods in this class can be used to run certain tests
    for a device as part of a test method (i.e. a method that starts with the
    string `test`).

    """

    batch_number: int
    eeprom_status: EEPROMStatus
    firmware_version: Version
    gtin: int
    hardware_version: Version
    oem_data: str
    operating_time: int
    power_off_cycles: int
    power_on_cycles: int
    production_date: date
    product_name: str
    release_name: str
    serial_number: str
    under_voltage_counter: int
    watchdog_reset_counter: int

    possible_attributes: List[SimpleNamespace] = []

    @classmethod
    def setUpClass(cls):
        """Set up data for whole test"""

        # Add basic test attributes that all devices share
        cls.add_attribute("EEPROM Status", "{cls.eeprom_status}", pdf=False)
        cls.add_attribute("Name", "{cls.name}")
        cls.add_attribute("Status", "{cls.status}")
        cls.add_attribute(
            "Production Date", "{cls.production_date}", pdf=False
        )
        cls.add_attribute("GTIN", "{cls.gtin}", pdf=False)
        cls.add_attribute("Product Name", "{cls.product_name}", pdf=False)
        cls.add_attribute("Batch Number", "{cls.batch_number}", pdf=False)
        cls.add_attribute("Bluetooth Address", "{cls.bluetooth_mac}")
        cls.add_attribute("RSSI", "{cls.bluetooth_rssi} dBm")
        cls.add_attribute("Hardware Version", "{cls.hardware_version}")
        cls.add_attribute("Firmware Version", "{cls.firmware_version}")
        cls.add_attribute("Release Name", "{cls.release_name}", pdf=False)
        cls.add_attribute("OEM Data", "{cls.oem_data}", pdf=False)
        cls.add_attribute(
            "Power On Cycles", "{cls.power_on_cycles}", pdf=False
        )
        cls.add_attribute(
            "Power Off Cycles", "{cls.power_off_cycles}", pdf=False
        )
        cls.add_attribute(
            "Under Voltage Counter", "{cls.under_voltage_counter}", pdf=False
        )
        cls.add_attribute(
            "Watchdog Reset Counter", "{cls.watchdog_reset_counter}", pdf=False
        )
        cls.add_attribute(
            "Operating Time", "{cls.operating_time} s", pdf=False
        )

        # Add a basic PDF report
        # Subclasses should overwrite this attribute, if you want to change
        # the default arguments of the report class
        cls.report = Report()

        # We store attributes related to the connection, such as MAC address
        # only once. To do that we set `read_attributes` to true after
        # the test class gathered the relevant data.
        cls.read_attributes = False

    @classmethod
    def tearDownClass(cls):
        """Print attributes of tested STH after all successful test cases"""

        cls.__output_general_data()
        cls.__output_node_data()
        cls.report.build()

    @classmethod
    def __output_general_data(cls):
        """Print general information and add it to PDF report"""

        now = datetime.now()
        date = now.strftime("%Y-%m-%d")
        time = now.strftime("%H:%M:%S")

        operator = settings.operator.name

        attributes = [
            SimpleNamespace(
                description="ICOc Version", value=__version__, pdf=True
            ),
            SimpleNamespace(description="Date", value=date, pdf=True),
            SimpleNamespace(description="Time", value=time, pdf=True),
            SimpleNamespace(description="Operator", value=operator, pdf=True),
        ]

        cls.__output_data(attributes, node_data=False)

    @classmethod
    def __output_node_data(cls):
        """Print node information and add it to PDF report"""

        attributes = []
        for attribute in cls.possible_attributes:
            try:
                attribute.value = str(attribute.value).format(cls=cls)
                attributes.append(attribute)
            except (AttributeError, IndexError):
                pass

        cls.__output_data(attributes)

    @classmethod
    def __output_data(cls, attributes, node_data=True):
        """Output data to standard output and PDF report

        Parameters
        ----------

        attributes:
            An iterable that stores simple name space objects created via
            ``create_attribute``

        node_data:
            Specifies if this method outputs node specific or general data
        """

        # Only output something, if there is at least one attribute
        if not attributes:
            return

        max_length_description = max(
            [len(attribute.description) for attribute in attributes]
        )
        max_length_value = max(
            [len(attribute.value) for attribute in attributes]
        )

        # Print attributes to standard output
        print("\n")
        header = "Attributes" if node_data else "General"
        print(header)
        print("—" * len(header))

        for attribute in attributes:
            print(
                f"{attribute.description:{max_length_description}} "
                + f"{attribute.value:>{max_length_value}}"
            )

        # Add attributes to PDF
        attributes_pdf = [
            attribute for attribute in attributes if attribute.pdf
        ]
        for attribute in attributes_pdf:
            cls.report.add_attribute(
                attribute.description, attribute.value, node_data
            )

    @classmethod
    def add_attribute(cls, name: str, value: object, pdf: bool = True) -> None:
        """Add a test attribute

        Parameters
        ----------

        name:
            The description (name) of the attribute

        value:
            The value of the attribute

        pdf:
            True if the attribute should be added to the PDF report

        """

        cls.possible_attributes.append(
            SimpleNamespace(description=name, value=str(value), pdf=pdf)
        )

    def setUp(self):
        """Set up hardware before a single test case"""

        # All tests methods that contain the text `disconnected` do not
        # initialize a Bluetooth connection
        if self._testMethodName.find("disconnected") >= 0:
            return

        self._connect()

        if not type(self).read_attributes:
            # Read data of specific node (STH or STU). Subclasses must
            # implement this method.
            self._read_data()
            type(self).read_attributes = True

    def tearDown(self):
        """Clean up after single test case"""

        # All tests methods that contain the text `disconnected` do not
        # initialize a Bluetooth connection
        if self._testMethodName.find("disconnected") >= 0:
            return

        self._disconnect()

    def run(self, result=None):
        """Execute a single test

        We override this method to store data about the test outcome.
        """

        super().run(result)
        type(self).report.add_test_result(self.shortDescription(), result)

    def _connect(self):
        """Create a connection to the STU"""

        async def connect():
            """Create connection with new network class"""

            self.can = Network()
            await self.can.reset_node("STU 1")
            # Wait for reset to take place
            await async_sleep(2)

        loop = new_event_loop()
        set_event_loop(loop)
        self.loop = loop
        self.loop.run_until_complete(connect())

    def _disconnect(self):
        """Tear down connection to STU"""

        self.loop.run_until_complete(self.can.shutdown())
        self.loop.close()

    async def _test_connection(self, node: str):
        """Check connection to node

        Parameters
        ----------

        node:
            The node for which the connection should be checked

        """

        # The sensor devices need a little more time to switch from the
        # “Startup” to the “Operating” state
        await async_sleep(1)

        # Just send a request for the state and check if the result matches
        # our expectations. The identifier of the answer will be checked by
        # (the notifier in) the network class already.
        state = await self.can.get_state(node)

        expected_state = State(
            mode="Get", location="Application", state="Operating"
        )

        self.assertEqual(
            expected_state,
            state,
            (
                f"Expected state “{expected_state}” does not match "
                f"received state “{state}”"
            ),
        )

    def _test_firmware_flash(
        self,
        node: str,
        flash_location: Union[str, Path],
        programmmer_serial_number: int,
        chip: str,
    ):
        """Upload bootloader and application into node

        Parameters
        ----------

        node:
            A textual description of the device/node that will be flashed

        flash_location:
            The location of the flash image

        programmer_serial_number:
            The serial number of the programming board

        chip:
            The name of the chip that should be flashed

        """

        image_filepath = Path(flash_location).expanduser().resolve()
        self.assertTrue(
            image_filepath.is_file(),
            f"Firmware file {image_filepath} does not exist",
        )

        commander = Commander(
            serial_number=programmmer_serial_number, chip=chip
        )

        commander.upload_flash(image_filepath)

    async def _test_eeprom_product_data(
        self, node: Node, config: DynaBox
    ) -> None:
        """Test if reading and writing the product data EEPROM page works

        Parameters
        ----------

        node:
            The node to which this method writes to and reads from

        config
            A configuration object that stores the various product data
            attributes

        """

        cls = type(self)

        receiver = node

        # ========
        # = GTIN =
        # ========

        gtin = config.gtin
        await self.can.write_eeprom_gtin(gtin, receiver)
        cls.gtin = await self.can.read_eeprom_gtin(receiver)
        self.assertEqual(
            gtin,
            cls.gtin,
            f"Written GTIN “{gtin}” does not match read GTIN “{cls.gtin}”",
        )

        # ====================
        # = Hardware Version =
        # ====================

        hardware_version = config.hardware_version
        await self.can.write_eeprom_hardware_version(
            hardware_version, receiver
        )
        cls.hardware_version = await self.can.read_eeprom_hardware_version(
            receiver
        )
        self.assertEqual(
            hardware_version,
            f"{cls.hardware_version}",
            f"Written hardware version “{hardware_version}” does not "
            + f"match read hardware version “{cls.hardware_version}”",
        )

        # ====================
        # = Firmware Version =
        # ====================

        await self.can.write_eeprom_firmware_version(
            cls.firmware_version, receiver
        )
        firmware_version = await self.can.read_eeprom_firmware_version(
            receiver
        )
        self.assertEqual(
            f"{cls.firmware_version}",
            f"{firmware_version}",
            f"Written firmware version “{cls.firmware_version}” does not "
            + f"match read firmware version “{firmware_version}”",
        )

        # ================
        # = Release Name =
        # ================

        # Originally we assumed that this value would be set by the firmware
        # itself. However, according to tests with an empty EEPROM this is not
        # the case.
        release_name = config.firmware.release_name
        await self.can.write_eeprom_release_name(release_name, receiver)
        cls.release_name = await self.can.read_eeprom_release_name(receiver)
        self.assertEqual(
            release_name,
            cls.release_name,
            f"Written firmware release name “{release_name}” does not "
            + f"match read firmware release name “{cls.release_name}”",
        )

        # =================
        # = Serial Number =
        # =================

        serial_number = config.serial_number
        await self.can.write_eeprom_serial_number(serial_number, receiver)
        cls.serial_number = await self.can.read_eeprom_serial_number(receiver)
        self.assertEqual(
            serial_number,
            cls.serial_number,
            f"Written serial number “{serial_number}” does not "
            + f"match read serial number “{cls.serial_number}”",
        )

        # ================
        # = Product Name =
        # ================

        product_name = config.product_name
        await self.can.write_eeprom_product_name(product_name, receiver)
        cls.product_name = await self.can.read_eeprom_product_name(receiver)
        self.assertEqual(
            product_name,
            cls.product_name,
            f"Written product name “{product_name}” does not "
            + f"match read product name “{cls.product_name}”",
        )

        # ============
        # = OEM Data =
        # ============

        oem_data = config.oem_data
        await self.can.write_eeprom_oem_data(oem_data, receiver)
        oem_data_list = await self.can.read_eeprom_oem_data(receiver)
        self.assertListEqual(
            oem_data,
            oem_data_list,
            f"Written OEM data “{oem_data}” does not "
            + f"match read OEM data “{oem_data_list}”",
        )
        # We currently store the data in text format, to improve the
        # readability of null bytes in the shell. Please notice, that this will
        # not always work (depending on the binary data stored in EEPROM
        # region).
        cls.oem_data = "".join(map(chr, oem_data_list)).replace("\x00", "")

    async def _test_eeprom_statistics(
        self, node: Node, production_date: date, batch_number: int
    ) -> None:
        """Test if reading and writing the statistics EEPROM page works

        For this purpose this method writes (default) values into the EEPROM,
        reads them and then checks if the written and read values are equal.

        Parameters
        ----------

        node:
            The device where the EEPROM statistics page should be updated

        production_date:
            The production date of the node

        batch_number:
            The batch number of the node

        """

        cls = type(self)
        receiver = repr(node)

        # =======================
        # = Power On/Off Cycles =
        # =======================

        power_on_cycles = 0
        await self.can.write_eeprom_power_on_cycles(power_on_cycles, receiver)
        cls.power_on_cycles = await self.can.read_eeprom_power_on_cycles(
            receiver
        )
        self.assertEqual(
            power_on_cycles,
            cls.power_on_cycles,
            f"Written power on cycle value “{power_on_cycles}” "
            + "does not match read power on cycle value "
            + f"“{cls.power_on_cycles}”",
        )

        power_off_cycles = 0
        await self.can.write_eeprom_power_off_cycles(
            power_off_cycles, receiver
        )
        cls.power_off_cycles = await self.can.read_eeprom_power_off_cycles(
            receiver
        )
        self.assertEqual(
            power_off_cycles,
            cls.power_off_cycles,
            f"Written power off cycle value “{power_off_cycles}” "
            + "does not match read power off cycle value "
            + f"“{cls.power_off_cycles}”",
        )

        # ==================
        # = Operating Time =
        # ==================

        operating_time = 0
        await self.can.write_eeprom_operating_time(operating_time, receiver)
        cls.operating_time = await self.can.read_eeprom_operating_time(
            receiver
        )
        self.assertEqual(
            operating_time,
            cls.operating_time,
            f"Written operating time “{operating_time}” "
            + "does not match read operating time “{cls.operating_time}”",
        )

        # =========================
        # = Under Voltage Counter =
        # =========================

        under_voltage_counter = 0
        await self.can.write_eeprom_under_voltage_counter(
            under_voltage_counter, receiver
        )
        cls.under_voltage_counter = (
            await self.can.read_eeprom_under_voltage_counter(receiver)
        )
        self.assertEqual(
            under_voltage_counter,
            cls.under_voltage_counter,
            f"Written under voltage counter value “{under_voltage_counter}” "
            + "does not match read under voltage counter value "
            + f"“{cls.under_voltage_counter}”",
        )

        # ==========================
        # = Watchdog Reset Counter =
        # ==========================

        watchdog_reset_counter = 0
        await self.can.write_eeprom_watchdog_reset_counter(
            watchdog_reset_counter, receiver
        )
        cls.watchdog_reset_counter = (
            await self.can.read_eeprom_watchdog_reset_counter(receiver)
        )
        self.assertEqual(
            watchdog_reset_counter,
            cls.watchdog_reset_counter,
            (
                "Written watchdog reset counter value"
                f" “{watchdog_reset_counter} does not match read watchdog"
                f" reset counter value “{cls.watchdog_reset_counter}”"
            ),
        )

        # ===================
        # = Production Date =
        # ===================

        await self.can.write_eeprom_production_date(production_date, receiver)
        cls.production_date = await self.can.read_eeprom_production_date(
            receiver
        )
        self.assertEqual(
            production_date,
            cls.production_date,
            f"Written production date “{production_date}” does not match "
            + f"read production date “{cls.production_date}”",
        )

        # ================
        # = Batch Number =
        # ================

        await self.can.write_eeprom_batch_number(batch_number, receiver)
        cls.batch_number = await self.can.read_eeprom_batch_number(receiver)
        self.assertEqual(
            batch_number,
            cls.batch_number,
            f"Written batch “{batch_number}” does not match "
            + f"read batch number “{cls.batch_number}”",
        )

    async def _test_eeprom_status(self, node: Node) -> None:
        """Test if reading and writing the EEPROM status byte works

        Attributes
        ----------

        node:
            The node where the status byte should be checked

        """

        cls = type(self)

        # =================
        # = EEPROM Status =
        # =================

        await self.can.write_eeprom_status("Initialized", node)
        cls.eeprom_status = await self.can.read_eeprom_status(node)
        self.assertTrue(
            cls.eeprom_status.is_initialized(),
            (
                "Setting EEPROM status to “Initialized” failed. "
                "EEPROM status byte currently stores the value "
                f"“{cls.eeprom_status}”"
            ),
        )



"""Experiment with some functionality of the „new“ Network class"""

# -- Imports ------------------------------------------------------------------

from asyncio import run
from functools import partial
from pathlib import Path
from time import time

from mytoolit.can import Network
from mytoolit.can.adc import ADCConfiguration
from mytoolit.measurement import convert_raw_to_g, Storage


# -- Functions ----------------------------------------------------------------


async def test(identifier):
    async with Network() as network:
        node = "STH 1"

        await network.connect_sensor_device(identifier)
        name = await network.get_name(node)
        mac_address = await network.get_mac_address(node)
        print(
            f"Connected to sensor device “{name}” with MAC "
            f"address “{mac_address}”"
        )

        # Reduce sample rate to decrease CPU usage
        # See: https://github.com/MyTooliT/ICOc/issues/40
        adc_config = ADCConfiguration(
            reference_voltage=3.3,
            prescaler=16,
            acquisition_time=8,
            oversampling_rate=64,
        )
        sample_rate = adc_config.sample_rate()
        print(f"Set sample rate to {sample_rate} Hz")
        await network.write_adc_configuration(**adc_config)

        sensor_range = await network.read_acceleration_sensor_range_in_g()
        print(f"Sensor Range: {sensor_range/2} g")

        conversion_to_g = partial(convert_raw_to_g, max_value=sensor_range)

        measurement_time_s = 10

        print(f"Measure acceleration values for {measurement_time_s} seconds…")
        filepath = Path(__file__).parent.parent.parent / "experiment.hdf5"
        if filepath.is_file():
            filepath.unlink()  # Remove old HDF file
        with Storage(filepath) as storage:
            start_time = time()
            async with network.open_data_stream(
                first=True, third=True
            ) as stream:
                async for data in stream:
                    data.apply(conversion_to_g)
                    storage.add_streaming_data(data)

                    if time() - start_time >= measurement_time_s:
                        break
            storage.add_acceleration_meta(
                "Sensor_Range", f"± {sensor_range} g₀"
            )


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    run(test(identifier=0))
